@@ -4,8 +4,13 @@ import string, inspect, sys, itertools
 from chainstuf import chainstuf  
 from options import Options, OptionsContext
 import six
+if six.PY3:
+    from codecs import getencoder
 
-# Define convenience functions
+### Convenience and compatibility functions and names
+
+stdout = sys.stdout
+stderr = sys.stderr
 
 def is_string(v):
     """
@@ -14,6 +19,41 @@ def is_string(v):
     """
     return isinstance(v, six.string_types)
 
+def encoded(u, encoding):
+    """
+    Encode string u (denoting it is expected to be in Unicode) if there's
+    encoding to be done. Tries to mask the difference between Python 2 and 3,
+    which have highly different models of string processing, and different
+    codec APIs and quirks. Some Python 3 encoders further require ``bytes``
+    in, not ``str``. These are first encoded into utf-8, encoded, then
+    decoded.
+    """
+    if not encoding:
+        return u
+    elif six.PY3:
+        try:
+            return getencoder(encoding)(u)[0]
+        except LookupError:
+            name = encoding.replace('-','')  # base-64 becomes base64 e.g.
+            bytesout = getencoder(name + '_codec')(u.encode('utf-8'))[0]
+            if name in set(['base64', 'hex', 'quopri', 'uu']):
+                return bytesout.decode('utf-8')
+            else:
+                return bytesout
+            
+        # NB PY3 requires lower-level interface for many codecs. s.encode('utf-8')
+        # works fine, but others do not. Some codecs convert bytes to bytes,
+        # and are not properly looked up by nickname (e.g. 'base64'). These are
+        # managed by first encoding into utf-8, then if it makes sense decoding
+        # back into a string. The others are things like bz2 and zlib--binary
+        # encodings that have little use to us here.
+        
+        # There are also be some slight variations in results that will make
+        # testing more fun. Possibly related to adding or not adding a terminal
+        # newline.
+    else:
+        return u.encode(encoding)
+        
 def flatten(*args):
     """
     Like itertools.chain(), but will pretend that single scalar values are singleton lists.
@@ -25,7 +65,8 @@ def flatten(*args):
     # would use ``hasattr(x, '__iter__')`` rather than ``isinstance(x, (list, tuple))``,
     # but other objects like file have ``__iter__``, which screws things up
 
-# Workhorse functions
+
+### Workhorse functions
 
 sformatter = string.Formatter()  # piggyback Python's format() template parser
 
@@ -46,15 +87,14 @@ def _sprintf(arg, caller, override=None):
             return eval(s, caller.f_globals, localvars)
         except SyntaxError:
             raise SyntaxError("syntax error when formatting '{}'".format(s))
-        except StandardError as se:
-            raise se
+
 
     if is_string(arg):
         parts = []
         for (literal_text, field_name, format_spec, conversion) in sformatter.parse(arg):
             parts.append(literal_text)
             if field_name is not None:
-                format_str = six.u("{") + ("!" + conversion if conversion else "") + \
+                format_str = six.u("{0") + ("!" + conversion if conversion else "") + \
                                           (":" + format_spec if format_spec else "") + "}"
                 field_value = seval(field_name)
                 formatted = format_str.format(field_value)
@@ -63,22 +103,35 @@ def _sprintf(arg, caller, override=None):
     else:
         return str(seval(str(arg)))
 
+### Core Say class
+
+DEFAULT_FILES = [ stdout]
 
 class Say(object):
-    
     """
     Say encapsulates printing functions. Instances are configurable, and callable.
     """
     
     options = Options(  
-        encoding='utf-8',   # character encoding for output
+        indent=0,           # indent level (if set to None, indentation is turned off)
+        indent_str='    ',  # indent string for each level
+        encoding=None if six.PY3 else 'utf-8',
+                            # character encoding for output (needed on Python 2, not 3)
         encoded=None,       # character encoding to return
-        file=sys.stdout,    # where is output headed? a write() able object, or str
+        files=None,         # where is output headed? a list of write() able objects
+                            # NB Set to [stdout] but in way that gets around stuf bug
         sep=' ',            # separate args with this (Python print function compatible)
         end='\n',           # end output with this (Python print function compatible)
         silent=False,       # do the formatting and return the result, but don't write the output
+        retvalue=False,     # should return the value of the formatted string?
         _callframe=None,    # frome from which the caller was calling
     )
+    options.files = DEFAULT_FILES # to get around a bug in stuf
+    
+    options.magic(
+        indent = lambda v, cur: cur.indent + int(v) if isinstance(v, str) else v
+    )
+    
   
     def __init__(self, **kwargs):
         """
@@ -125,25 +178,8 @@ class Say(object):
         An update-like function.
         """
         self.options.set(**kwargs)
-    
-    def settings(self, **kwargs):
-        """
-        Open a context manager for a `with` statement. Temporarily change settings
-        for the duration of the with.
-        """
-        return SayContext(self, kwargs)
-    
-    def clone(self, **kwargs):
-        """
-        Create a new Say instance whose options are chained to this instance's
-        options (and thence to Say.options). bkwargs become the cloned instance's
-        overlay options.
-        """
-        cloned = Say()
-        cloned.options = self.options.push(kwargs)
-        return cloned
-    
-    def setfiles(self, *files):
+        
+    def setfiles(self, files):
         """
         Set the list of output files. ``files`` is a list. For each item, if
         it's a real file like ``sys.stdout``, use it. If it's a string, assume
@@ -155,9 +191,28 @@ class Say(object):
             Else, assume it's an open file. Just return it.
             """
             return open(f, "w") if is_string(f) else f
-
-        self.options.file = [ opened(f) for f in flatten(*files) ]
+            
+        self.options.files = [ opened(f) for f in files ]
         return self
+    
+        # TBD: Turn this into 'magical' attribute set
+    
+    def settings(self, **kwargs):
+        """
+        Open a context manager for a `with` statement. Temporarily change settings
+        for the duration of the with.
+        """
+        return SayContext(self, kwargs)
+    
+    def clone(self, **kwargs):
+        """
+        Create a new Say instance whose options are chained to this instance's
+        options (and thence to Say.options). kwargs become the cloned instance's
+        overlay options.
+        """
+        cloned = Say()
+        cloned.options = self.options.push(kwargs)
+        return cloned
 
     def __call__(self, *args, **kwargs): 
         
@@ -166,47 +221,44 @@ class Say(object):
         
         formatted = [ _sprintf(arg, opts._callframe) if is_string(arg) else str(arg)
                       for arg in args ]
-        return self._output_str(opts.sep.join(formatted), opts)
+        return self._output(opts.sep.join(formatted), opts)
         
-    def _output_str(self, outstr, opts):
+    def _output(self, data, opts):
         """
-        Do the actual formatting and outputting work. 
+        Do the actual formatting and outputting work. ``data`` may be either a
+        list of lines, or a composed string.
         """
         
-        def encoded(s, encoding):
-            """
-            Encode the strings if they're encoding to be done.
-            """
-            return s.encode(encoding) if encoding else s
+        if opts.indent:
+            indent_str = opts.indent_str * opts.indent
+            datalines = data if isinstance(data, list) else data.splitlines()
+            outstr = '\n'.join([ ''.join([ indent_str, line ]) for line in datalines ])
+        else:
+            outstr = '\n'.join(data) if isinstance(data, list) else data
+        # by end of indenting, dealing with string only
 
         # prepare and emit output
         if opts.end is not None:
             outstr += opts.end
         if not opts.silent:
-            self._write(encoded(outstr, opts.encoding), opts)
+            to_write = encoded(outstr, opts.encoding)
+            self._write(to_write, opts)
             
         # prepare and return return value
-        retencoding = opts.encoding if opts.encoded is True else opts.encoded
-        retencoded = encoded(outstr, retencoding)
-        rettrimmed = retencoded[:-1] if retencoded.endswith('\n') else retencoded
-        return rettrimmed
-        
-    def _output(self, lines, opts):
-        """
-        Alternate output function that accepts list of lines, not composed strings.
-        
-        More important and logical in the future. Stay tuned.
-        """
-        self._output_str("\n".join(lines), opts)
-    
+        if opts.retvalue:
+            retencoding = opts.encoding if opts.encoded is True else opts.encoded
+            retencoded = encoded(outstr, retencoding)
+            rettrimmed = retencoded[:-1] if retencoded.endswith('\n') else retencoded
+            return rettrimmed
+
     def _write(self, s, opts):
         """
         Write s to all associated file objects. 
         """
-        for f in flatten(opts.file):
+        # print(opts)
+        for f in opts.files:
             f.write(s)
 
-
 class SayContext(OptionsContext):
     """
     Context helper to support Python's with statement.  Generally called
@@ -214,5 +266,10 @@ class SayContext(OptionsContext):
     """
     pass
     
+
+### Define default ``say`` and ``fmt`` callables
+
 say = Say()
-fmt = say.clone(silent=True).setfiles()
+
+fmt = say.clone(encoding=None, retvalue=True, silent=True)
+fmt.setfiles([])
