@@ -1,5 +1,4 @@
 
-
 import six, os
 from say import Say, fmt, stdout
 
@@ -129,7 +128,15 @@ def test_indent():
     assert say('force no indent', indent=0) == 'force no indent'
     say.set(indent=0)
     assert say('no indent again') == 'no indent again'
-
+    
+def test_sep_end():
+    setup()
+    assert say(1, 2, 3) == '1 2 3'
+    assert say(1, 2, 3, sep=',') == '1,2,3'
+    assert say(1, 2, 3, sep=', ') == '1, 2, 3'
+    say('xyz', end='\n') == 'xyz\n'
+    say('xyz', end='\n\n') == 'xyz\n\n'
+    say('xyz', end='') == 'xyz'
 
 def test_indent_special():
     setup()
